#include <stdio.h>
#include <stdlib.h>
#include <time.h>

unsigned long long get_cycles() {
	unsigned int lo, hi;
	asm volatile (
		"\tcpuid\n"
		"\trdtsc\n"
		: "=a"(lo), "=d"(hi) /* outputs =a/=d means use eax/edx*/
   		:                    /* inputs (no inputs) */
   		: "%ebx", "%ecx"     /* clobbers*/

	);
	unsigned long long ret_val = ((unsigned long long)hi) << 32 | lo;
	return ret_val;
}

int count_bits_c(unsigned int input) {
	int bits = 0;
	while (input != 0) {
		if (input & 1) bits++;
		input >>= 1;
	}
	return bits;
}

int count_bits_asm1(unsigned int input) {
	int bits;

	asm volatile (
		"\t  xorl %%eax, %%eax  # Clear the eax register\n"
		"\t1:\n"
		"\t  testl %%ebx, %%ebx # Check to see if ebx in 0\n"
		"\t  je 3f              # End the loop, if it is\n"
		"\t  movl %%ebx, %%ecx  # Move ebx into temp register\n"
		"\t  andl $1, %%ecx     # Check the low order bit (this sets zf if 0)\n"
		"\t  jz 2f              # If the low order bit is 0, just shift\n"
		"\t  incl %%eax         # Otherwise, incr bits\n"
		"\t2:\n"
		"\t  shrl %%ebx         # Shift the input right. Use logical since we want 0's\n"
		"\t  jmp 1b             # Continue the loop\n"
		"\t3:\n"

		: "=a"(bits)         /* outputs =a  use eax */
   		: "b"(input)         /* inputs =b put input into ebx */
   		: "%ecx"             /* clobbers */
	);

	return bits;
}

int count_bits_asm2(unsigned int input) {
	int bits;

	asm volatile (
		"\t  xorl %%eax, %%eax   # Clear the eax register\n"
		"\t  popcnt %%ebx, %%eax # Counts the number of bits\n"

		: "=a"(bits)         /* outputs =a  use eax */
   		: "b"(input)         /* inputs =b put input into ebx */
   		:                    /* clobbers */
	);

	return bits;
}

int count_bits_builtin(unsigned int input) {
	return __builtin_popcount(input);
}

int main(int argc, char *argv[]) {
	if (argc != 2) {
		printf("USAGE: %s <tests>\n", argv[0]);
		return -1;
	}

	int tests = atoi(argv[1]);
	if (tests < 1) {
		printf("Number must be >1\n");
		printf("USAGE: %s <tests>\n", argv[0]);
		return -1;
	}
	srand(time(NULL));

	unsigned int *array = (unsigned int*)malloc(sizeof(unsigned int)*tests);
	for (int i=0; i<tests; i++) {
		array[i] = rand();
	}

	if (tests == 1) {
		printf("Using value %d\n", array[0]);
		printf("Using count_bits_c got %d\n", count_bits_c(array[0]));
		printf("Using count_bits_asm1 got %d\n", count_bits_asm1(array[0]));
		printf("Using count_bits_asm2 got %d\n", count_bits_asm2(array[0]));
		printf("Using count_bits_builtin got %d\n", count_bits_builtin(array[0]));
		return 0;
	}

	unsigned int total_bits;

	unsigned long long start, end;

	total_bits = 0;
	start = get_cycles();
	for (int i=0; i<tests; i++) {
		total_bits += count_bits_c(array[i]);
	}
	end = get_cycles();
	printf("Using count_bits_c took       %llu cycles\n", end-start);
	printf("Got %u bits\n", total_bits);

	total_bits = 0;
	start = get_cycles();
	for (int i=0; i<tests; i++) {
		total_bits += count_bits_asm1(array[i]);
	}
	end = get_cycles();
	printf("Using count_bits_asm1 took    %llu cycles\n", end-start);
	printf("Got %u bits\n", total_bits);

	total_bits = 0;
	start = get_cycles();
	for (int i=0; i<tests; i++) {
		total_bits += count_bits_asm2(array[i]);
	}
	end = get_cycles();
	printf("Using count_bits_asm2 took    %llu cycles\n", end-start);
	printf("Got %u bits\n", total_bits);

	total_bits = 0;
	start = get_cycles();
	for (int i=0; i<tests; i++) {
		total_bits += count_bits_builtin(array[i]);
	}
	end = get_cycles();
	printf("Using count_bits_builtin took %llu cycles\n", end-start);
	printf("Got %u bits\n", total_bits);

	return 0;
}